"""
Implements different clock / PLL models for easy manipulation without vendor software.
"""

import json
import re
import numpy as np
#from . import utils
import os
import fractions
from collections import defaultdict

class EnumVal:
    def __init__(self, name, value, description):
        self.name = name
        self.value = value
        self.description = description
        self.__doc__ = self.description

    @property
    def __pydoc__(self):
        return self.description

    def get(self):
        return self.value

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

class ConstantField:
    def __init__(self, field):
        self.end = field["end"]
        self.start = field["start"]
        self.value = field["value"]
        self.name = "CONST"
        self.description = ""

        self.width = self.end - self.start + 1
        self.mask = ((1 << self.width) - 1) << self.start

    def get(self):
        return self.value

    def get_raw(self):
        return self.mask & (self.value << self.start)

    def parse(self, obj):
        pass

    @property
    def value_description(self):
        return ""

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return str(self.value)

class Field:
    def __init__(self, field):
        self.end = field["end"]
        self.start = field["start"]
        self.width = self.end - self.start + 1
        self.mask = ((1 << self.width) - 1) << self.start

        self.name = field["name"]
        self.description = field["description"]
        self.__doc__ = self.description
        self.default = field["default"] if "default" in field else 0
        self.value = self.default

        self.enum_map = {}

        valid = field["valid"]
        self.valid_type = valid["type"]

        if self.valid_type == "int":
            pass # ??
        elif self.valid_type == "constant":
            self.default = valid["value"]
            self.value = valid["value"]
        elif self.valid_type == "enum":
            for value in valid["values"]:
                enum_val = EnumVal(**value)
                assert getattr(self, value["name"], None) is None, f"Duplicate enum value in field {self.name}: {value['name']}"
                setattr(self, value["name"], enum_val)
                self.enum_map[enum_val.value] = enum_val
        else:
            raise RuntimeError("Unknown valid type: " + self.valid_type)

    def get(self):
        if self.valid_type == "enum":
            if self.value in self.enum_map:
                return self.enum_map[self.value]
            else:
                return "BAD ENUM VALUE"
        return self.value

    def set(self, value):
        if self.value_type == "enum":
            if not isinstance(value, EnumVal):
                raise RuntimeError("Expected enum value!")
            else:
                self.value = value.value
        else:
            self.value = value

    def get_raw(self):
        return self.mask & (self.value << self.start)

    def parse(self, data):
        self.value = (data & self.mask) >> self.start

    def set(self, val):
        if isinstance(val, int):
            self.value = val
        elif isinstance(val, EnumVal):
            self.value = val.value
        else:
            raise RuntimeError("Unsupported type!")

    def reset(self):
        self.value = self.default

    @property
    def value_description(self):
        if self.valid_type == "enum":
            if self.value in self.enum_map:
                return self.enum_map[self.value].description
            else:
                return "BAD ENUM VALUE"
        return ""

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return str(self.__dict__)

class Register:
    def __init__(self, obj, dw=8):
        self.addr = obj["addr"]
        self.fields = []
        self.dw = dw 
        self.regdef = obj

        for field in self.regdef["fields"]:
            fieldtype = field["fieldtype"]
            if fieldtype == "constant":
                self.fields.append(ConstantField(field))
            elif fieldtype == "normal":
                newfield = Field(field)
                newfield.index = len(self.fields)
                self.fields.append(newfield)
            else:
                raise RuntimeError("Unsupported field type!")

    def reset(self):
        for field in self.fields:
            field.reset()

    def parse(self, val):
        for field in self.fields:
            field.parse(val)

    def __str__(self):
        return str({"addr": self.addr, "fields": self.fields})

    def __repr__(self):
        return str({"addr": self.addr, "fields": self.fields})

    def get_raw(self):
        ret = self.addr << self.dw

        for field in self.fields:
            ret |= field.get_raw()

        return ret

class MultiRegister:
    def __init__(self, parent, name, fields):
        self.parent = parent
        self.fields = fields
        self.name = name

    @property
    def value(self):
        return self.parent.get_long_register(*self.fields)

    @value.setter
    def value(self, v):
        self.parent.set_long_register(v, *self.fields)

class RegisterDevice:
    def __init__(self, aw, dw, definition):
        # Address width and data width
        self.aw = aw
        self.dw = dw
        self.registers_by_addr = {}
        self.regname_pattern = re.compile(r"[A-Za-z0-9_]+\[(\d+):(\d+)\]")

        defpath = os.path.join(os.path.dirname(__file__), definition)
        with open(defpath) as f:
            regmap = json.load(f)

        multi_regs = {}

        for register in regmap:
            addr = register["addr"]
            reg = Register(register, dw=dw)
            self.registers_by_addr[addr] = reg

            for field in reg.fields:
                field.addr = addr
                if isinstance(field, Field) and field.valid_type != "constant":
                    sanitized_name = field.name.replace("[", "_").replace("]", "").replace(":", "_")
                    if field.name.endswith("]"): # Multi-field
                        name = field.name[:field.name.index("[")]
                        multi_regs[name] = multi_regs.get(name, []) + [field]
                    setattr(self, sanitized_name, field)

        for k,v in multi_regs.items():
            setattr(self, k, MultiRegister(self, k, v))

    def init_from_file(self, file):
        if hasattr(file, "read"):
            lines = file.read().strip().split("\n")
        else:
            with open(file, "r") as f:
                lines = f.read().strip().split("\n")

        for line in lines:
            a,b = line.split("\t")
            rawdata = int(b, 16)

            addr_mask = (1 << self.aw) - 1
            addr = addr_mask & (rawdata >> self.dw)
            data_mask = (1 << self.dw) - 1
            data = rawdata & data_mask

            if not addr in self.registers_by_addr:
                print(f"Unhandled register: {hex(addr)}, skipping ...")
                print()
                continue

            self.registers_by_addr[addr].parse(data)

        self.update()

    def get_long_register(self, *args):
        ret = 0
        for field in args:
            match = self.regname_pattern.match(field.name)
            if match is None:
                raise RuntimeError("Cannot read non-long field: " + field.name)

            end = int(match.group(1))
            start = int(match.group(2))
            width = end-start+1
            mask = ((1 << width)-1)

            val = (field.value & mask)
            ret |= val << start

        return ret

    def set_long_register(self, value, *args):
        for field in args:
            match = self.regname_pattern.match(field.name)
            if match is None:
                raise RuntimeError("Cannot read non-long field: " + field.name)

            end = int(match.group(1))
            start = int(match.group(2))
            width = end-start+1
            mask = ((1 << width)-1)
            value_mask = mask << start

            field.value = (value & value_mask) >> start

    def get_register_dump(self, with_addr=False):
        ret = []
        for addr in self.register_addresses:
            if not with_addr:
                ret.append(self.registers_by_addr[addr].get_raw())
            else:
                ret.append((addr, self.registers_by_addr[addr].get_raw()))

        return ret

    def print(self):
        for addr in self.register_addresses:
            register = self.registers_by_addr[addr]

            for field in register.fields:
                val = field.get()
                if field.name == "CONST":
                    continue

                print(f"    {field.name}")
                print(f"        Description: {field.description}")
                print(f"        Value:       {val}")
                if field.valid_type == "enum":
                    print(f"        ValDesc:     {field.value_description}")
                print()

LMX2594_VCOs = [
        (1,  7500,  8600, 164, 12, 299, 240),
        (2,  8600,  9800, 165, 16, 356, 247),
        (3,  9800, 10800, 158, 19, 324, 224),
        (4, 10800, 12000, 140,  0, 383, 244),
        (5, 12000, 12900, 183, 36, 205, 146),
        (6, 12900, 13900, 155,  6, 242, 163),
        (7, 13900, 15000, 175, 19, 323, 244)
    ]

CAL_NO_ASSIST = 0
CAL_PARTIAL_ASSIST = 1
CAL_CLOSE_FREQUENCY_ASSIST = 2
CAL_FULL_ASSIST = 3

CHDIV_TABLE = [
        (2, 15000, 3750, 7500),
        (4, 15000, 1875, 3750),
        (6, 15000, 1250, 2500),
        (8, 11500, 937.5, 1437.5),
        (12, 11500, 625, 958.333),
        (16, 11500, 468.75, 718.75),
        (24, 11500, 312.5, 479.167),
        (32, 11500, 234.375, 359.375),
        (48, 11500, 156.25, 239.583),
        (64, 11500, 117.1875, 179.6875),
        (72, 11500, 104.167, 159.722),
        (96, 11500, 78.125, 119.792),
        (128, 11500, 58.594, 89.844),
        (192, 11500, 39.0625, 59.896),
        (256, 11500, 29.297, 44.922),
        (384, 11500, 19.531, 29.948),
        (512, 11500, 14.648, 22.461),
        (768, 11500, 9.766, 14.974),
        (1, 15000, 10, 15000)
        ]

class LMX2594(RegisterDevice):

    def __init__(self, f_osc):
        RegisterDevice.__init__(self, 8, 16, "lmx2594_regmap.json")

        self.f_osc = f_osc

        # Note that this isn't the complete range, but skips the readback registers
        self.register_addresses = list(range(109, -1, -1))

    def get_multiplier_freqs(self):
        """
        Enumerate all possible multiplier output frequencies.
        """
        f2conf = defaultdict(list)
        # config tuple order is mult, osc_x, r_pre - we want this to be sortable
        # we prefer smaller mult first, then smaller osc_x
        # add no-multiplier configs (there are no limits on using r_pre in this mode, but you usually don't need it)
        f2conf[self.f_osc].append((1,1,1))
        f2conf[2*self.f_osc].append((1,2,1))
        for osc_x in [1,2]:
            if osc_x==2 and self.f_osc>200: # max input freq of doubler
                continue
            for mult in range(3,8):
                # apply limits on the multiplier's input and output freqs
                min_r_pre = int(np.ceil(max(1, self.f_osc*osc_x/70.0, self.f_osc*osc_x*mult/250.0)))
                max_r_pre = int(np.floor(min(255, self.f_osc*osc_x/30.0, self.f_osc*osc_x*mult/180.0)))
                for r_pre in range(min_r_pre, max_r_pre+1):
                    mult_out = self.f_osc*(osc_x*mult/r_pre)
                    f2conf[mult_out].append((mult, osc_x, r_pre))

        f2conf = dict(sorted(f2conf.items()))
        for freq in f2conf.keys():
            f2conf[freq].sort()
            #for conf in f2conf[freq]:
            #    print(freq, conf)
            f2conf[freq] = f2conf[freq][0]
        return f2conf

    def set_output_frequency(self, f_target, pwr=31, solution=None, en_b=False, osc_2x=False, verbose=True):
        # We only support integer mode right now
        self.MASH_ORDER.value = 0
        self.MASH_RESET_N.set(self.MASH_RESET_N.RESET)
        self.PLL_NUM.value = 0
        self.PLL_DEN.value = 0

        self.OUTA_PWR.value = pwr
        self.OUTA_PD.set(self.OUTA_PD.NORMAL_OPERATION)
        if en_b:
            self.OUTB_PWR.value = pwr
            self.OUTB_PD.set(self.OUTB_PD.NORMAL_OPERATION)
        else:
            self.OUTB_PWR.value = 0
            self.OUTB_PD.set(self.OUTB_PD.POWERDOWN)

        chdivs = []
        f_vco_min = 7500

        mult = 1
        osc_x = 2 if osc_2x else 1
        for i,(div,f_vco_max,f_out_min,f_out_max) in enumerate(CHDIV_TABLE):
            if not (f_out_min <= f_target <= f_out_max):
                continue

            f_vco = f_target * div

            if not (f_vco_min <= f_vco <= f_vco_max):
                continue

            chdivs.append((i, div, f_vco))

        if len(chdivs) < 1:
            raise RuntimeError("No possible integer solutions found!")

        solutions = []
        if verbose:
            print("  i |   f_vco  | DIV | DLY_SEL |   n  | osc_2x |   R  | mult | R_pre |  f_pfd  |   f_out  | Delta f |   Metric   ")
            print("----|----------|-----|---------|------|--------|------|------|-------|---------|----------|---------|------------")

        metric_min = np.inf
        metric_min_idx = None
        for idx,(i,div,f_vco) in enumerate(chdivs):
            min_n,dly_sel = LMX2594.get_modulator_constraints(self.MASH_ORDER.value, f_vco)

            ratio = fractions.Fraction(f_vco / (self.f_osc*osc_x)).limit_denominator(255)
            n = ratio.numerator
            R = ratio.denominator

            R_pre = 1

            assert n != 0, "N can't be zero!"

            while n < min_n or self.f_osc / (R * R_pre) > 400:
                n *= 2

                if R*2 <= 255:
                    R *= 2
                elif R_pre * 2 < 128:
                    R_pre *= 2
                else:
                    raise RuntimeError("Failed to find solution, N limit can't be met!")

            f_pd = self.f_osc * osc_x / (R * R_pre)

            metric = R_pre * R * n * div
            if metric < metric_min:
                metric_min_idx = idx
                metric_min = metric

            f_out = f_vco / div
            delta_f = abs(f_out - f_target)

            metric += delta_f*1e6

            if verbose:
                print(f" {idx:>2d} | {f_vco:8.2f} | {div:3d} | {dly_sel:7d} | {n:4d} | {'  True' if osc_x==2 else ' False'} | {R:4d} | {mult:4d} | {R_pre:5d} | {f_pd:7.2f} | {f_out:8.2f} | {delta_f:7.2f} | {metric:6.4e}")

            solutions.append((i, div, f_vco, dly_sel, n, osc_x, R, mult, R_pre))

        if verbose: print()
        if solution is None:
            if verbose: print(f"Choosing solution {metric_min_idx} with minimal metric {metric_min}.")
            solution = metric_min_idx

        chdiv_i,chdiv,f_vco,dly_sel,n,osc_x,R,mult,R_pre = solutions[solution]

        self.CHDIV.value = chdiv_i % 18
        self.PFD_DLY_SEL.value = dly_sel
        self.PLL_N.value = n
        self.OSC_2X.value = osc_x - 1
        self.PLL_R_PRE.value = R_pre
        self.MULT.value = mult
        self.PLL_R.value = R

        if chdiv_i == 18:
            self.OUTA_MUX.set(self.OUTA_MUX.VCO)
            self.OUTB_MUX.set(self.OUTB_MUX.VCO)
            self.CHDIV_DIV2.set(self.CHDIV_DIV2.DISABLED)
        else:
            self.OUTA_MUX.set(self.OUTA_MUX.CHANNEL_DIVIDER)
            self.OUTB_MUX.set(self.OUTB_MUX.CHANNEL_DIVIDER)

            # Enable CHDIV_DIV2 driver for CHDIV > 2
            self.CHDIV_DIV2.set(self.CHDIV_DIV2.DISABLED if chdiv_i == 0 else self.CHDIV_DIV2.ENABLED)

        self.update()
        self.configure_calibration()

        return f_vco / chdiv

    def configure_calibration(self, assistance_level=0):
        if self.f_pd <= 100:
            if self.f_pd >= 10:
                self.FCAL_LPFD_ADJ.value = 0
            elif self.f_pd >= 5:
                self.FCAL_LPFD_ADJ.value = 1
            elif self.f_pd >= 2.5:
                self.FCAL_LPFD_ADJ.value = 2
            else:
                self.FCAL_LPFD_ADJ.value = 3

            self.FCAL_HPFD_ADJ.value = 0
        elif self.f_pd <= 150:
            self.FCAL_HPFD_ADJ.value = 1
        elif self.f_pd <= 200:
            self.FCAL_HPFD_ADJ.value = 2
        else:
            self.FCAL_HPFD_ADJ.value = 3

        # Optimize for phase noise performance (At the cost of locking time)
        self.CAL_CLK_DIV.value = 3
        self.ACAL_CMP_DLY.value = 25

        # Don't output clock signal during calibration
        self.OUT_MUTE.value = 1
        self.OUT_FORCE.value = 0

        # VCO calibration settings
        if 11900 <= self.f_vco <= 12100:
            self.VCO_SEL.value = 4
            self.QUICK_RECAL_EN.value = 0
            self.VCO_SEL_FORCE.value = 0
            self.VCO_DACISET_FORCE.value = 0
            self.VCO_CAPCTRL_FORCE.value = 0

            self.VCO_DACISET_STRT.value = 300
            self.VCO_CAPCTRL_STRT.value = 183 

        elif assistance_level == CAL_NO_ASSIST:
            self.QUICK_RECAL_EN.value = 0
            self.VCO_SEL_FORCE.value = 0
            self.VCO_DACISET_FORCE.value = 0
            self.VCO_CAPCTRL_FORCE.value = 0
            self.VCO_SEL.value = 7

        else:
            for vco_id, f_min, f_max, c_min, c_max, a_min, a_max in LMX2594_VCOs:
                if f_min <= self.f_vco <= f_max:
                    self.VCO_SEL.value = vco_id
                    self.VCO_DACISET_STRT.value = round(c_min - (c_min - c_max) * (f_vco - f_min) / (f_max - f_min))
                    self.VCO_CAPCTRL_STRT.value = round(a_min + (a_max - a_min) * (f_vco - f_min) / (f_max - f_min))
                    break

                if vco_id == 7:
                    raise RuntimeError("Failed to find acceptable VCO??")

            if assistance_level == CAL_PARTIAL_ASSIST:
                self.QUICK_RECAL_EN.value = 0
                self.VCO_SEL_FORCE.value = 0
                self.VCO_DACISET_FORCE.value = 0
                self.VCO_CAPCTRL_FORCE.value = 0

            elif assistance_level == CAL_CLOSE_FREQUENCY_ASSIST:
                self.QUICK_RECAL_EN.value = 1
                self.VCO_SEL_FORCE.value = 0
                self.VCO_DACISET_FORCE.value = 0
                self.VCO_CAPCTRL_FORCE.value = 0

            elif assistance_level == CAL_FULL_ASSIST:
                self.QUICK_RECAL_EN.value = 0
                self.VCO_SEL_FORCE.value = 1
                self.VCO_DACISET_FORCE.value = 1
                self.VCO_CAPCTRL_FORCE.value = 1

    def get_modulator_constraints(mash_order, f_vco):
        if mash_order == 0: # for now, it's always 0
            if f_vco <= 12500:
                min_n = 28
                dly_sel = 1
            else:
                min_n = 32
                dly_sel = 2
        elif mash_order == 1:
            if f_vco <= 10000:
                min_n = 28
                dly_sel = 1
            elif 10000 < f_vco < 12500:
                min_n = 32
                dly_sel = 2
            else:
                min_n = 36
                dly_sel = 3
        elif mash_order == 2:
            if f_vco <= 10000:
                min_n = 32
                dly_sel = 2
            else:
                min_n = 36
                dly_sel = 3
        elif mash_order == 3:
            if f_vco <= 10000:
                min_n = 36
                dly_sel = 3
            else:
                min_n = 40
                dly_sel = 4
        elif mash_order == 4:
            if f_vco <= 10000:
                min_n = 44
                dly_sel = 5
            else:
                min_n = 48
                dly_sel = 6
        else:
            raise RuntimeError(f"Can't handle unknown mash order: {mash_order}")

        return min_n, dly_sel

    def update(self):
        """
        Compute the output frequencies from the register values.
        """
        if self.OSC_2X.get() == self.OSC_2X.ENABLED:
            f_in = 2 * self.f_osc
        else:
            f_in = self.f_osc

        f_in /= self.PLL_R_PRE.value
        f_in *= self.MULT.value
        f_in /= self.PLL_R.value

        self.f_pd = f_in

        num = self.PLL_NUM.value
        den = self.PLL_DEN.value
        pll_n = self.PLL_N.value

        if den != 0 and num != 0:
            self.f_vco = self.f_pd * (pll_n + num/den)
        else:
            self.f_vco = self.f_pd * pll_n

        if not (7500 <= self.f_vco <= 15000):
            raise RuntimeError(f"VCO frequency f_vco = {round(self.f_vco, ndigits=2)} out of range: 7500 MHz <= f_vco <= 15000")

        self.f_chdiv = self.f_vco / CHDIV_TABLE[self.CHDIV.value][0]

        if self.OUTA_MUX.get() == self.OUTA_MUX.CHANNEL_DIVIDER:
            self.f_outa = self.f_chdiv
        else:
            self.f_outa = self.f_vco

        sysref_divider_lut = { self.SYSREF_DIV_PRE.DIVIDE_BY_1: 1,
                               self.SYSREF_DIV_PRE.DIVIDE_BY_2: 2,
                               self.SYSREF_DIV_PRE.DIVIDE_BY_4: 4 }
        
        if self.SYSREF_EN.get() == self.SYSREF_EN.ENABLED:
            sysref_div = self.SYSREF_DIV.get()
            sysref_div *= sysref_divider_lut[self.SYSREF_DIV_PRE.get()]
            self.f_sysref = self.f_vco / sysref_div

        outb_mux = self.OUTB_MUX.get()
        if outb_mux == self.OUTB_MUX.CHANNEL_DIVIDER:
            self.f_outb = self.f_chdiv
        elif outb_mux == self.OUTB_MUX.VCO or outb_mux == self.OUTB_MUX.HIGH_IMPEDANCE:
            self.f_outb = self.f_vco
        elif outb_mux == self.SYSREF:
            self.f_outb = self.f_sysref

class LMK04828BOutputBranch:
    def __init__(self, parent, i):
        self.parent = parent
        self.i = i

        def ga(n):
            return getattr(self.parent, n)

        self.DIV = ga(f"DCLKout{self.i}_DIV")
        self.CLK_PD = ga(f"CLKout{self.i}_{self.i+1}_PD")
        self.SDCLK_PD = ga(f"SDCLKout{self.i+1}_PD")
        self.DCLK_FMT = ga(f"DCLKout{self.i}_FMT")
        self.SDCLK_FMT = ga(f"SDCLKout{self.i+1}_FMT")
        self.DCLK_POL = ga(f"DCLKout{self.i}_POL")
        self.SDCLK_POL = ga(f"SDCLKout{self.i+1}_POL")

        self.DCLK_MUX = ga(f"DCLKout{self.i}_MUX")
        self.SDCLK_MUX = ga(f"SDCLKout{self.i+1}_MUX")

    def _set_output_status(self, dclk_enable, sdclk_enable):
        if not dclk_enable and not sdclk_enable:
            self.CLK_PD.set(self.CLK_PD.POWERDOWN)
        elif dclk_enable and not sdclk_enable:
            self.CLK_PD.set(self.CLK_PD.ENABLED)
            self.DCLK_FMT.set(self.DCLK_FMT.LVDS)
            self.SDCLK_PD.set(self.SDCLK_PD.POWERDOWN)
        elif not dclk_enable and sdclk_enable:
            self.CLK_PD.set(self.CLK_PD.ENABLED)
            self.DCLK_FMT.set(self.DCLK_FMT.POWERDOWN)
            self.SDCLK_FMT.set(self.SDCLK_FMT.LVDS)
            self.SDCLK_PD.set(self.SDCLK_PD.ENABLED)
        elif dclk_enable and sdclk_enable:
            self.CLK_PD.set(self.CLK_PD.ENABLED)
            self.DCLK_FMT.set(self.DCLK_FMT.LVDS)
            self.SDCLK_FMT.set(self.SDCLK_FMT.LVDS)
            self.SDCLK_PD.set(self.SDCLK_PD.ENABLED)
        else:
            raise RuntimeError("?!")

    @property
    def dclk_active(self):
        return not (self.dclk_fmt == self.DCLK_FMT.POWERDOWN or self.clk_pd == self.CLK_PD.POWERDOWN)

    @dclk_active.setter
    def dclk_active(self, value: bool):
        if self.dclk_active == value:
            return

        self._set_output_status(value, self.sdclk_active)
        self.parent.update()

    @property
    def sdclk_active(self):
        return not (self.sdclk_pd == self.SDCLK_PD.POWERDOWN or self.sdclk_fmt == self.SDCLK_FMT.POWERDOWN or self.clk_pd == self.CLK_PD.POWERDOWN)

    @sdclk_active.setter
    def sdclk_active(self, value: bool):
        if self.sdclk_active == value:
            return

        self._set_output_status(self.dclk_active, value)
        self.parent.update()

    def get_sdclk_freqs(self):
        return self.parent.pll2_output_freq / (np.arange(32)+1)

    def request_freq(self, value, ignore_warning=False):
        _div = self.parent.pll2_output_freq / value
        div = max(1, min(32, round(_div)))

        f_real = self.parent.pll2_output_freq / div

        if abs(div - _div) > 0.01 and not ignore_warning:
            print(f"WARNING: Failed to hit requested frequency of {value:.2f} MHz. Using divider {div} where {_div:.3f} would have been required, thus resulting in output frequency {f_real:.2f} MHz or a frequency error of {f_real - value:.2f} MHz!")

        self.DIV.value = 0x1f & div # 32 maps to 0
        self.update()

        return self.parent.pll2_output_freq / div

    def update(self, printDebug=False):
        def dbg(*s):
            if printDebug:
                print(f"Output Branch {self.i:2d}", *s)

        self.div = self.DIV.get()
        if self.div == 0: # 0 means 32
            self.div = 32

        dbg(f"DIV: {self.div}")

        self.clk_pd = self.CLK_PD.get()
        dbg("CLK_PD:", self.clk_pd)

        self.sdclk_pd = self.SDCLK_PD.get()
        dbg("SDCLK_PD:", self.sdclk_pd)

        self.dclk_fmt = self.DCLK_FMT.get()
        dbg("DCLK_FMT:", self.dclk_fmt)

        self.sdclk_fmt = self.SDCLK_FMT.get()
        dbg("SDCLK_FMT:", self.sdclk_fmt)

        self.dclk_pol = self.DCLK_POL.get()
        dbg("DCLK_POL:", self.dclk_pol)

        self.sdclk_pol = self.SDCLK_POL.get()
        dbg("SDCLK_POL:", self.sdclk_pol)

        self.sdclk_mux = self.SDCLK_MUX.get()
        dbg("SDCLK_MUX:", self.sdclk_mux)

        self.dclk_mux = self.DCLK_MUX.get()
        dbg("DCLK_MUX:", self.dclk_mux)

        pll_freq = self.parent.pll2_output_freq
        sysref_freq = self.parent.sysref_freq

        if self.dclk_mux == self.DCLK_MUX.BYPASS:
            self.dclk_freq = pll_freq
        else:
            self.dclk_freq = pll_freq / self.div

        if self.sdclk_mux == self.SDCLK_MUX.DEVICE_CLOCK_OUTPUT:
            self.sdclk_freq = self.dclk_freq
        else:
            self.sdclk_freq = sysref_freq

        dbg("DCLK_FREQ:", round(self.dclk_freq, ndigits=2))
        dbg("SDCLK_FREQ:", round(self.sdclk_freq, ndigits=2))

        dbg("DCLK_ACTIVE:", self.dclk_active)
        dbg("SDCLK_ACTIVE:", self.sdclk_active)

class LMK04828B(RegisterDevice):
    def __init__(self, clkin0_freq, clkin1_freq, clkin2_freq, vcxo_freq):
        RegisterDevice.__init__(self, 13, 8, "data/lmk04828b_regmap.json")

        self.clkin0_freq = clkin0_freq
        self.clkin1_freq = clkin1_freq
        self.clkin2_freq = clkin2_freq
        self.vcxo_freq = vcxo_freq

        # Dictionaries are unordered, and because the order of operations is important when writing
        # these registers, the correct order of indices is written down in this array:
        self.register_addresses = [0, 2, 3, 4, 5, 6, 12, 13, 256, 257, 258, 259, 260, 261, 262,
                                   263, 264, 265, 266, 267, 268, 269, 270, 271, 272, 273, 274, 275,
                                   276, 277, 278, 279, 280, 281, 282, 283, 284, 285, 286, 287, 288,
                                   289, 290, 291, 292, 293, 294, 295, 296, 297, 298, 299, 300, 301,
                                   302, 303, 304, 305, 306, 307, 308, 309, 310, 311, 312, 313, 314,
                                   315, 316, 317, 318, 319, 320, 321, 322, 323, 324, 325, 326, 327,
                                   328, 329, 330, 331, 332, 333, 334, 335, 336, 337, 338, 339, 340,
                                   341, 342, 343, 344, 345, 346, 347, 348, 349, 350, 351, 352, 353,
                                   354, 355, 356, 357, 369, 370, 380, 381, 358, 359, 360, 361, 362,
                                   363, 364, 365, 366, 371, 386, 387, 388, 389, 392, 393, 394, 395,
                                   8189, 8190, 8191]

        self.clock_branches = [LMK04828BOutputBranch(self, 2*i) for i in range(7)]

    def write_register_dump(self, name):
        with open(name, "w") as f:
            f.write(f"R0 (INIT)\t0x000090\n")

            for addr,val in self.get_register_dump(with_addr=True):
                f.write(f"R{addr}\t0x{val:06X}\n")

    def update(self, printDebug=False):
        def dbg(*s):
            if printDebug:
                print(*s)

        sel_mode = self.CLKin_SEL_MODE.get()

        dbg("SEL_MODE:", self.CLKin_SEL_MODE.get())

        if sel_mode == self.CLKin_SEL_MODE.CLK_IN_0_MANUAL:
            assert self.CLKin0_OUT_MUX.get() == self.CLKin0_OUT_MUX.PLL1
            pll1_src_freq = self.clkin0_freq

            divider = self.CLKin0_R.value

        elif sel_mode == self.CLKin_SEL_MODE.CLK_IN_1_MANUAL:
            assert self.CLKin1_OUT_MUX.get() == self.CLKin1_OUT_MUX.PLL1
            pll1_src_freq = self.clkin1_freq

            divider = self.CLKin1_R.value

        elif sel_mode == self.CLKin_SEL_MODE.CLK_IN_2_MANUAL:
            assert self.CLKin2_OUT_MUX.get() == self.CLKin2_OUT_MUX.PLL1
            pll1_src_freq = self.clkin2_freq

            divider = self.CLKin2_R.value
        else:
            raise RuntimeError("sel_mode == " + str(sel_mode) + ", which is not supported!")

        dbg("Input divider:", divider)

        assert self.PLL1_NCLK_MUX.get() == self.PLL1_NCLK_MUX.OSC_IN, "Only configurations using the external VCXO are supported!"

        self.pll1_phase_detector_freq = pll1_src_freq / divider
        dbg("PLL1 Phase Detector Frequency:", self.pll1_phase_detector_freq)

        self.pll1_n_divider = self.PLL1_N.value
        dbg("PLL1N Divider:", self.pll1_n_divider)
        dbg("Expected VCXO Frequency:", self.pll1_n_divider * self.pll1_phase_detector_freq)

        assert self.vcxo_freq == self.pll1_n_divider * self.pll1_phase_detector_freq

        self.pll2_r_divider = self.PLL2_R.value
        pll2_input_freq = self.vcxo_freq / self.pll2_r_divider * (2 if self.PLL2_REF_2X_EN.get() == self.PLL2_REF_2X_EN.ENABLED else 1)
        dbg("PLL2 Input Frequency:", pll2_input_freq)

        assert self.PLL2_NCLK_MUX.get() == self.PLL2_NCLK_MUX.PLL_PRESCALER, "Only configurations using PLL2 feedback from prescaler are supported!"

        pll2_p_raw = self.PLL2_P.get()
        if pll2_p_raw == self.PLL2_P.DIVIDE_2 or pll2_p_raw == self.PLL2_P.DIVIDE_2_2:
            pll2_p = 2
        elif pll2_p_raw == self.PLL2_P.DIVIDE_8:
            pll2_p = 8
        else:
            pll2_p = self.PLL2_P.value

        dbg("PLL2_P:", pll2_p)

        pll2_n = self.PLL2_N.value
        dbg("PLL2_N:", pll2_n)

        self.pll2_output_freq = pll2_input_freq * pll2_n * pll2_p
        dbg("PLL2 Output Frequency:", round(self.pll2_output_freq, ndigits=2))

        self.sysref_divider = self.SYSREF_DIV.value
        dbg("SYSREF DIVIDER:", self.sysref_divider)

        self.sysref_freq = self.pll2_output_freq / self.sysref_divider
        dbg("SYSREF FREQ:", round(self.sysref_freq, ndigits=5))

        for branch in self.clock_branches:
            branch.update(printDebug)

    def set_refclk(self, refclk, precision=1):
        f_i = int(self.vcxo_freq * 10**precision)

        if 2370 < refclk < 2630:
            vco = 0 # VCO0
        elif 2920 < refclk < 3080:
            vco = 1 # VCO1
        else:
            raise RuntimeError("Requested VCO frequency is not compatible with either VCO")

        refclk = int(refclk * 10**precision)

        div = np.gcd(f_i, refclk)

        R = f_i // div
        N = refclk // div

        if self.vcxo_freq // R > 150:
            print("Phase detector frequency too high, introducing factor")
            fac = np.ceil((self.vcxo_freq // R) / 150)
            R *= fac
            N *= fac

        success = False
        for P in range(2, 9): # Absorb part of N into the prescaler
            if N % P == 0:
                success = True
                break

        if not success:
            raise RuntimeError("Failed to find suitable solution!")

        self.VCO_MUX.value = self.VCO_MUX.VCO_1.value if vco == 1 else self.VCO_MUX.VCO_0.value

        self.PLL2_R.value = R
        self.PLL2_N.value = N // P
        self.PLL2_P.value = P & 0x7 # 8 is represented as 0

        self.update()

        return R, N // P, P

    def set_sysref(self, value):
        _div = self.pll2_output_freq / value
        div = min(8191, max(8, int(round(_div))))

        f_new = self.pll2_output_freq / div

        if abs(div - _div) / div > 0.01:
            print(f"WARNING: SYSREF_CLK target could not be hit accurately! Requested frequency {value} MHz requires divider {_div:.4f} which is not realizable. The closest integer divider {div} results in a frequency of {f_new} MHz!")

        self.SYSREF_DIV.value = div
        self.update()

class CLK104Output:
    def __init__(self, branch):
        self.branch = branch

    @property
    def freq(self):
        return self.branch.dclk_freq

    @freq.setter
    def freq(self, value):
        return self.branch.request_freq(value)

    @property
    def sysref_freq(self):
        return self.branch.sdclk_freq

    @property
    def enable(self):
        return self.branch.dclk_active

    @property
    def sysref_enable(self):
        return self.branch.sdclk_active

    @enable.setter
    def enable(self, value):
        self.branch.dclk_active = value

    @sysref_enable.setter
    def sysref_enable(self, value):
        self.branch.sdclk_active = value

class CLK104:
    def __init__(self, src=None):
        # 10 MHz reference clock, 10 MHz clock input on external SMA, 160 MHz VCO frequency
        self.lmk = LMK04828B(10, 10, 10, 160)
        self.lmx_adc = LMX2594(245.76)
        self.lmx_dac = LMX2594(245.76)

        if src is None:
            with files("ipq_pynq_utils").joinpath("data/lmk04828b_regdump_defaults.txt").open() as f:
                self.lmk.init_from_file(f)
        else:
            self.lmk.init_from_file(src)

        with files("ipq_pynq_utils").joinpath("data/clockFiles/LMX2594_REF-245M76__OUT-9830M40_10172019_I.txt").open() as f:
            self.lmx_adc.init_from_file(f)

        with files("ipq_pynq_utils").joinpath("data/clockFiles/LMX2594_REF-245M76__OUT-9830M40_10172019_I.txt").open() as f:
            self.lmx_dac.init_from_file(f)

        self.RF_PLL_ADC_REF = CLK104Output(self.lmk.clock_branches[0])
        self.AMS_SYSREF = CLK104Output(self.lmk.clock_branches[1])
        self.RF_PLL_DAC_REF = CLK104Output(self.lmk.clock_branches[2])
        self.DAC_REFCLK = CLK104Output(self.lmk.clock_branches[3])
        self.PL_CLK = CLK104Output(self.lmk.clock_branches[4])
        self.EXT_REF_OUT = CLK104Output(self.lmk.clock_branches[5])
        self.ADC_REFCLK = CLK104Output(self.lmk.clock_branches[6])

    @property
    def PLL2_FREQ(self):
        return self.lmk.pll2_output_freq

    @PLL2_FREQ.setter
    def PLL2_FREQ(self, value):
        self.lmk.set_refclk(value)

    @property
    def SYSREF_FREQ(self):
        return self.lmk.sysref_freq

    @SYSREF_FREQ.setter
    def SYSREF_FREQ(self, value):
        self.lmk.set_sysref(value)

    def get_register_dump(self):
        return {
                "LMK": self.lmk.get_register_dump()
                }



"""
Drivers for signal generators: FPGA blocks that send data to DACs.
"""
from pynq.buffer import allocate
import numpy as np
from qick import SocIp

class AbsSignalGen(SocIp):
    """
    Abstract class which defines methods that are common to different signal generators.
    """
    # The DAC channel has a mixer.
    HAS_MIXER = False
    # Waveform samples per fabric clock.
    SAMPS_PER_CLK = 1
    # Maximum waveform amplitude.
    MAXV = 2**15-2
    # Scale factor between MAXV and the default maximum amplitude (necessary to avoid overshoot).
    MAXV_SCALE = 1.0

    # Configure this driver with links to the other drivers, and the signal gen channel number.
    def configure(self, ch, rf):
        # Channel number corresponding to entry in the QickConfig list of gens.
        self.ch = ch

        # RF data converter
        self.rf = rf

        self.cfg['dac'] = self.dac
        self.cfg['has_mixer'] = self.HAS_MIXER

        for p in ['fs', 'fs_mult', 'fs_div', 'interpolation', 'f_fabric']:
            self.cfg[p] = self.rf.daccfg[self['dac']][p]
        # interpolation reduces the DDS range
        self.cfg['f_dds'] = self['fs']/self['interpolation']
        self.cfg['fdds_div'] = self['fs_div']*self['interpolation']

    def configure_connections(self, soc):
        self.soc = soc

        # what RFDC port does this generator drive?
        block, port, _ = soc.metadata.trace_forward(self['fullpath'], 'm_axis', ["usp_rf_data_converter"])
        # port names are of the form 's00_axis'
        self.dac = port[1:3]

        #print("%s: switch %d, tProc ch %d, DAC tile %s block %s"%(self.fullpath, self.switch_ch, self.tproc_ch, *self.dac))

    def set_nyquist(self, nqz):
        """Set the Nyquist zone mode for the DAC linked to this generator.
        For tProc-controlled generators, this method is called automatically during program config.
        You should normally only call this method directly for a constant-IQ output.

        Parameters
        ----------
        nqz : int
            Nyquist zone (must be 1 or 2).
            Setting the NQZ to 2 increases output power in the 2nd/3rd Nyquist zones.
        """
        self.rf.set_nyquist(self.dac, nqz)

    def set_mixer_freq(self, f, ro_ch=None):
        """Set the mixer frequency for the DAC linked to this generator.
        For tProc-controlled generators, this method is called automatically during program config.
        You should normally only call this method directly for a constant-IQ output.

        Parameters
        ----------
        mixer_freq : float
            Mixer frequency (in MHz)
        ro_ch : int
            readout channel for frequency matching (use None if you don't want mixer freq to be rounded to a valid readout frequency)
        """
        if not self.HAS_MIXER:
            raise NotImplementedError("This channel does not have a mixer.")
        if ro_ch is None:
            self.rf.set_mixer_freq(self.dac, f)
        else:
            mixercfg = self.soc._get_mixer_cfg(self.ch)
            rocfg = self.soc['readouts'][ro_ch]
            rounded_f = self.soc.roundfreq(f, [mixercfg, rocfg])
            self.rf.set_mixer_freq(self.dac, rounded_f)

    def get_mixer_freq(self):
        if not self.HAS_MIXER:
            raise NotImplementedError("This channel does not have a mixer.")
        return self.rf.get_mixer_freq(self.dac)

class AbsArbSignalGen(AbsSignalGen):
    """
    A signal generator with a memory for envelope waveforms.
    """
    # Name of the input driven by the waveform DMA (if applicable).
    WAVEFORM_PORT = 's0_axis'

    def configure(self, ch, rf):
        # Define buffer.
        self.buff = allocate(shape=self.MAX_LENGTH, dtype=np.int32)

        super().configure(ch, rf)

    def configure_connections(self, soc):
        super().configure_connections(soc)

        # what switch port drives this generator?
        ((block, port),) = soc.metadata.trace_bus(self.fullpath, self.WAVEFORM_PORT)
        self.switch = getattr(soc, block)
        # port names are of the form 'M01_AXIS'
        self.switch_ch = int(port.split('_')[0][1:])
        ((block, port),) = soc.metadata.trace_bus(block, 'S00_AXIS')
        self.dma = getattr(soc, block)

    # Load waveforms.
    def load(self, xin, addr=0):
        """
        Load waveform into I,Q envelope

        :param xin: array of (I, Q) values for pulse envelope
        :type xin: int16 array
        :param addr: starting address
        :type addr: int
        """
        length = xin.shape[0]
        assert xin.dtype==np.int16

        # Check for max length.
        if length+addr > self.MAX_LENGTH:
            raise RuntimeError("%s: buffer length must be %d samples or less." %
                  (self.__class__.__name__, self.MAX_LENGTH))

        # Check for even transfer size.
        #if length % 2 != 0:
        #    raise RuntimeError("Buffer transfer length must be even number.")

        # Route switch to channel.
        self.switch.sel(mst=self.switch_ch)

        #print(self.fullpath, xin.shape, addr, self.switch_ch)

        # Pack the data into a single array; columns will be concatenated
        # -> lower 16 bits: I value.
        # -> higher 16 bits: Q value.
        # Format and copy data.
        np.copyto(self.buff[:length],
                np.frombuffer(xin, dtype=np.int32))

        ################
        ### Load I/Q ###
        ################
        # Enable writes.
        self._wr_enable(addr)

        # DMA data.
        self.dma.sendchannel.transfer(self.buff, nbytes=int(length*4))
        self.dma.sendchannel.wait()

        # Disable writes.
        self._wr_disable()

    def _wr_enable(self, addr=0):
        """
           Enable WE reg
        """
        self.start_addr_reg = addr
        self.we_reg = 1

    def _wr_disable(self):
        """
           Disable WE reg
        """
        self.we_reg = 0

class AbsPulsedSignalGen(AbsSignalGen):
    """
    A signal generator controlled by the TProcessor.
    """
    # Name of the input driven by the tProc (if applicable).
    TPROC_PORT = 's1_axis'
    B_PHASE = None

    def configure(self, ch, rf):
        super().configure(ch, rf)
        # DDS sampling frequency.
        self.cfg['maxlen'] = self.MAX_LENGTH
        self.cfg['b_dds'] = self.B_DDS
        if self.B_PHASE is not None: self.cfg['b_phase'] = self.B_PHASE
        self.cfg['switch_ch'] = self.switch_ch
        self.cfg['samps_per_clk'] = self.SAMPS_PER_CLK
        self.cfg['maxv'] = self.MAXV
        self.cfg['maxv_scale'] = self.MAXV_SCALE

    def configure_connections(self, soc):
        super().configure_connections(soc)

        # what tProc output port drives this generator?
        block, port, blocktype = soc.metadata.trace_back(self['fullpath'], self.TPROC_PORT, ["axis_tproc64x32_x8", "qick_processor", "axis_tmux_v1"])

        if blocktype == "axis_tmux_v1":
            # which tmux port drives this generator?
            # port names are of the form 'm2_axis'
            self.cfg['tmux_ch'] = int(port.split('_')[0][1:])
            ((block, port),) = soc.metadata.trace_bus(block, "s_axis")

        # ask the tproc to translate this port name to a channel number
        self.cfg['tproc_ch'],_ = getattr(soc, block).port2ch(port)

class AxisSignalGen(AbsArbSignalGen, AbsPulsedSignalGen):
    """
    AxisSignalGen class
    Supports AxisSignalGen V4+V5+V6, since they have the same software interface (ignoring registers that are not used)

    AXIS Signal Generator Registers.
    START_ADDR_REG

    WE_REG
    * 0 : disable writes.
    * 1 : enable writes.
    """
    bindto = ['user.org:user:axis_signal_gen_v4:1.0',
              'user.org:user:axis_signal_gen_v5:1.0',
              'user.org:user:axis_signal_gen_v6:1.0']
    REGISTERS = {'start_addr_reg': 0, 'we_reg': 1, 'rndq_reg': 2}
    SAMPS_PER_CLK = 16
    B_DDS = 32
    B_PHASE = 32

    def __init__(self, description):
        """
        Constructor method
        """
        super().__init__(description)

        # Default registers.
        self.start_addr_reg = 0
        self.we_reg = 0
        self.rndq_reg = 10

        # Generics
        self.N = int(description['parameters']['N'])
        self.NDDS = int(description['parameters']['N_DDS'])

        # Maximum number of samples
        self.MAX_LENGTH = 2**self.N*self.NDDS

    def rndq(self, sel_):
        """
           TODO: remove this function. This functionality was removed from IP block.
        """
        self.rndq_reg = sel_

class AxisSgInt4V1(AbsArbSignalGen, AbsPulsedSignalGen):
    """
    AxisSgInt4V1

    The default max amplitude for this generator is 0.9 times the maximum of int16.
    This is necessary to prevent interpolation overshoot:
    the output of the interpolation filter may exceed the max value of the input points.
    (https://blogs.keysight.com/blogs/tech/rfmw.entry.html/2019/05/07/confronting_measurem-IBRp.html)
    The result of overshoot is integer overflow in the filter output and big negative spikes.
    If the input to the filter is a square pulse, the rising edge of the output overshoots by 10%.
    Therefore, scaling envelopes by 90% seems safe.

    AXIS Signal Generator with envelope x4 interpolation V1 Registers.
    START_ADDR_REG

    WE_REG
    * 0 : disable writes.
    * 1 : enable writes.
    """
    bindto = ['user.org:user:axis_sg_int4_v1:1.0']
    REGISTERS = {'start_addr_reg': 0, 'we_reg': 1}
    HAS_MIXER = True
    FS_INTERPOLATION = 4
    MAXV_SCALE = 0.9
    B_DDS = 16
    B_PHASE = 16

    def __init__(self, description):
        """
        Constructor method
        """
        super().__init__(description)

        # Default registers.
        self.start_addr_reg = 0
        self.we_reg = 0

        # Generics
        self.N = int(description['parameters']['N'])
        self.NDDS = 4  # Fixed by design, not accesible.

        # Maximum number of samples
        # Table is interpolated. Length is given only by parameter N.
        self.MAX_LENGTH = 2**self.N


class AbsMuxSignalGen(AbsPulsedSignalGen):
    """
    Generic class for multiplexed generators.

    Registers:
    PINCx_REG : frequency of tone x.
    POFFx_REG : phase of tone x.
    GAINx_REG : gain of tone x.

    WE_REG
    * 0 : disable writes.
    * 1 : enable writes.
    """

    TPROC_PORT = 's_axis'
    # these must be defined by the subclass
    HAS_MIXER = None
    B_DDS = None
    N_TONES = None
    HAS_GAIN = None
    HAS_PHASE = None
    B_PHASE = None

    def __init__(self, description):
        """
        Constructor method
        """
        # define the register map
        iReg = 0
        for i in range(self.N_TONES): self.REGISTERS['pinc%d_reg'%(i)] = i + iReg
        iReg += self.N_TONES
        if self.HAS_PHASE:
            for i in range(self.N_TONES): self.REGISTERS['poff%d_reg'%(i)] = i + iReg
            iReg += self.N_TONES
        if self.HAS_GAIN:
            for i in range(self.N_TONES): self.REGISTERS['gain%d_reg'%(i)] = i + iReg
            iReg += self.N_TONES
        self.REGISTERS['we_reg'] = iReg

        super().__init__(description)

        self.cfg['n_tones'] = self.N_TONES
        self.cfg['has_gain'] = self.HAS_GAIN
        self.cfg['has_phase'] = self.HAS_PHASE

        # Generics
        self.NDDS = int(description['parameters']['N_DDS'])

        # dummy values, since this doesn't have a waveform memory.
        self.switch_ch = -1
        self.MAX_LENGTH = 0

        # Default registers.
        for i in range(self.N_TONES):
            setattr(self, 'pinc{}_reg'.format(i), 0)
            setattr(self, 'poff{}_reg'.format(i), 0)
            setattr(self, 'gain{}_reg'.format(i), self.MAXV)

        self.update()

    def update(self):
        """
        Update register values
        """
        self.we_reg = 1
        self.we_reg = 0

    def set_tones_int(self, tones):
        """Set up a list of tones all at once, using raw (integer) units.
        If the supplied list of tones is shorter than the number supported, the extra tones will have their gains set to 0.

        This method isn't meant to be called directly. It is called by set_tones() or QickProgram.config_gens().

        Parameters
        ----------
        tones : list of dict
            Tones to configure.
            The tone parameters are defined with keys freq_int, gain_int, phase_int.
            Omit parameters not supported by this version of the generator.
            All supported parameters must be defined.
        """
        if len(tones) > self.N_TONES:
            raise RuntimeError("Too many tones defined for this mux generator.")
        for i in range(self.N_TONES):
            if i < len(tones):
                tone = tones[i]
                setattr(self,'pinc%d_reg'%(i), tone['freq_int'])
                if self.HAS_GAIN:
                    setattr(self,'gain%d_reg'%(i), tone['gain_int'])
                if self.HAS_PHASE:
                    setattr(self,'poff%d_reg'%(i), tone['phase_int'])
            else:
                # zero the gain of unused tones
                if self.HAS_GAIN:
                    setattr(self,'gain%d_reg'%(i), 0)
        # Register update.
        self.update()

    def set_tones(self, freqs, gains=None, phases=None, ro_ch=None):
        """Set up a list of tones.

        This method is not normally used, it's only for debugging and testing.
        Normally the generator is configured based on parameters supplied in QickProgram.declare_gen().

        Parameters
        ----------
        freqs : list of float
            Tone frequencies for the muxed generator (in MHz).
            Positive and negative values are allowed.
        gains : list of float, optional
            Tone amplitudes for the muxed generator (in range -1 to 1).
        phases : list of float, optional
            Phases for the muxed generator (in degrees).
        ro_ch : int, optional
            readout channel for frequency-matching
        """
        tones = self.soc.calc_mux_regs(self.ch, freqs, gains, phases, ro_ch)
        self.set_all_int(tones)

class AxisSgMux4V1(AbsPulsedSignalGen):
    """
    AxisSgMux4V1

    AXIS Signal Generator with 4 muxed outputs.
    """
    bindto = ['user.org:user:axis_sg_mux4_v1:1.0']
    HAS_MIXER = True
    B_DDS = 16
    N_TONES = 4
    HAS_GAIN = False
    HAS_PHASE = False

class AxisSgMux4V2(AbsMuxSignalGen):
    """
    AxisSgMux4V2

    AXIS Signal Generator with 4 muxed outputs.
    """
    bindto = ['user.org:user:axis_sg_mux4_v2:1.0']
    HAS_MIXER = True
    B_DDS = 32
    N_TONES = 4
    HAS_GAIN = True
    HAS_PHASE = False

class AxisSgMux4V3(AxisSgMux4V2):
    """AxisSgMux4V3: no digital mixer, but otherwise behaves identically to AxisSgMux4V2.
    """
    bindto = ['user.org:user:axis_sg_mux4_v3:1.0']
    HAS_MIXER = False

class AxisSgMux8V1(AbsMuxSignalGen):
    """
    AxisSgMux8V1

    AXIS Signal Generator with 8 muxed outputs.
    """
    bindto = ['user.org:user:axis_sg_mux8_v1:1.0']
    HAS_MIXER = False
    B_DDS = 32
    N_TONES = 8
    HAS_GAIN = True
    HAS_PHASE = True
    B_PHASE = 32

class AxisConstantIQ(AbsSignalGen):
    """Plays a constant IQ value, which gets mixed with the DAC's built-in oscillator.
    """
    # AXIS Constant IQ registers:
    # REAL_REG : 16-bit.
    # IMAG_REG : 16-bit.
    # WE_REG   : 1-bit. Update registers.
    bindto = ['user.org:user:axis_constant_iq:1.0']
    REGISTERS = {'real_reg': 0, 'imag_reg': 1, 'we_reg': 2}
    HAS_MIXER = True

    def __init__(self, description):
        # Initialize ip
        super().__init__(description)

        # Default registers.
        self.real_reg = self.MAXV
        self.imag_reg = self.MAXV

        # Register update.
        self.update()

    def update(self):
        self.we_reg = 1
        self.we_reg = 0

    def set_iq(self, i=1, q=1):
        """
        Set gain.

        Parameters
        ----------
        i : float
            signed gain, I component (in range -1 to 1)
        q : float
            signed gain, Q component (in range -1 to 1)
        """
        # Set registers.
        self.real_reg = np.int16(i*self.MAXV)
        self.imag_reg = np.int16(q*self.MAXV)

        # Register update.
        self.update()



#!/usr/bin/env python3
# Typically (use the IP and port of your Pyro nameserver):
# python -m qick.pyro_cli myqick -n 192.168.133.17 -p 8888
import sys, getopt
from qick.pyro import start_server

ns_host = None
ns_port = None
proxy_name = "myqick"

options, remainder = getopt.gnu_getopt(sys.argv[1:], 'n:p:h')
for opt, arg in options:
    if opt == '-n':
        ns_host = arg
    elif opt == '-p':
        ns_port = int(arg)
    elif opt == '-h':
        print("\nUsage: "+sys.argv[0]+" [server name]")
        print("Must be run as root.")
        print("Arguments: ")
        print("\t-n: nameserver hostname (default localhost)")
        print("\t-p: nameserver port (default 9090)")
        print("\t-i: network interface (default eth0)")
        print("\t-h: this message")
        print("\n")
        print("On pynq 2.7 you may get an error relating to pynq not loading, or no XRT devices being found.")
        print("If this happens, try the following (taken from /usr/local/bin/start_jupyter.sh):")
        print("\tsudo -s")
        print("\tfor f in /etc/profile.d/*.sh; do source $f; done")
        print("\t"+sys.argv[0]+" [options]")
        print("\n")
        sys.exit(0)

if remainder:
    proxy_name = remainder[0]

start_server(ns_host, ns_port, proxy_name)



"""
Function to parse tProc assembly language programs.
"""
import re

# Function to parse program.


def parse_prog(file="prog.asm", outfmt="bin"):
    """
    Parses the .asm assembly language tProc program into a specified output format (binary or hex)

    :param file: ASM program file name
    :type file: str
    :param outfmt: Output format ("bin" or "hex")
    :type outfmt: str
    :return: Program in the new output format
    :rtype: bin or hex
    """
    # Output structure.
    outProg = {}

    # Instructions.
    instList = {}

    # I-type.
    instList['pushi'] = {'bin': '00010000'}
    instList['popi'] = {'bin': '00010001'}
    instList['mathi'] = {'bin': '00010010'}
    instList['seti'] = {'bin': '00010011'}
    instList['synci'] = {'bin': '00010100'}
    instList['waiti'] = {'bin': '00010101'}
    instList['bitwi'] = {'bin': '00010110'}
    instList['memri'] = {'bin': '00010111'}
    instList['memwi'] = {'bin': '00011000'}
    instList['regwi'] = {'bin': '00011001'}
    instList['setbi'] = {'bin': '00011010'}

    # J-type.
    instList['loopnz'] = {'bin': '00110000'}
    instList['condj'] = {'bin': '00110001'}
    instList['end'] = {'bin': '00111111'}

    # R-type.
    instList['math'] = {'bin': '01010000'}
    instList['set'] = {'bin': '01010001'}
    instList['sync'] = {'bin': '01010010'}
    instList['read'] = {'bin': '01010011'}
    instList['wait'] = {'bin': '01010100'}
    instList['bitw'] = {'bin': '01010101'}
    instList['memr'] = {'bin': '01010110'}
    instList['memw'] = {'bin': '01010111'}
    instList['setb'] = {'bin': '01011000'}

    # Structures for symbols and program.
    progList = {}
    symbList = {}

    ##############################
    ### Read program from file ###
    ##############################
    fd = open(file, "r")
    addr = 0
    for line in fd:
        # Match comments.
        m = re.search("^\s*//", line)

        # If there is a match.
        if m:
            # print(line)
            a = 1

        else:
            # Match instructions.
            jump_re = "^((.+):)?"
            inst_re_I = "pushi|popi|mathi|seti|synci|waiti|bitwi|memri|memwi|regwi|setbi|"
            inst_re_J = "loopnz|condj|end|"
            inst_re_R = "math|set|sync|read|wait|bitw|memr|memw|setb"
            inst_re = "\s*(" + inst_re_I + inst_re_J + inst_re_R + ")\s+(.+);"
            comp_re = jump_re + inst_re
            m = re.search(comp_re, line, flags=re.MULTILINE)

            # If there is a match.
            if m:
                # Tagged instruction for jump.
                if m.group(2):
                    symb = m.group(2)
                    inst = m.group(3)
                    args = m.group(4)

                    # Add symbol to symbList.
                    symbList[symb] = addr

                    # Add instruction to progList.
                    progList[addr] = {'inst': inst, 'args': args}

                    # Increment address.
                    addr = addr + 1

                # Normal instruction.
                else:
                    inst = m.group(3)
                    args = m.group(4)

                    # Add instruction to progList.
                    progList[addr] = {'inst': inst, 'args': args}

                    # Increment address.
                    addr = addr + 1

            # Check special case of "end" instruction.
            else:
                m = re.search("\s*(end)\s*;", line)

                # If there is a match.
                if m:
                    # Add instruction to progList.
                    progList[addr] = {'inst': 'end', 'args': ''}

                    # Increment address.
                    addr = addr + 1

    #########################
    ### Support functions ###
    #########################
    def unsigned2bin(strin, bits=8):
        maxv = 2**bits - 1

        # Check if hex string.
        m = re.search("^0x", strin, flags=re.MULTILINE)
        if m:
            dec = int(strin, 16)
        else:
            dec = int(strin, 10)

        # Check max.
        if dec > maxv:
            print("Error: number %d is bigger than %d" % (dec, maxv))
            return None

        # Convert to binary.
        fmt = "{0:0" + str(bits) + "b}"
        binv = fmt.format(dec)

        return binv

    def integer2bin(strin, bits=8):
        minv = -2**(bits-1)
        maxv = 2**(bits-1) - 1

        # Check if hex string.
        m = re.search("^0x", strin, flags=re.MULTILINE)
        if m:
            # Special case for hex number.
            dec = int(strin, 16)

            # Convert to binary.
            fmt = "{0:0" + str(bits) + "b}"
            binv = fmt.format(dec)

            return binv
        else:
            dec = int(strin, 10)

        # Check max.
        if dec < minv:
            print("Error: number %d is smaller than %d" % (dec, minv))
            return None

        # Check max.
        if dec > maxv:
            print("Error: number %d is bigger than %d" % (dec, maxv))
            return None

        # Check if number is negative.
        if dec < 0:
            dec = dec + 2**bits

        # Convert to binary.
        fmt = "{0:0" + str(bits) + "b}"
        binv = fmt.format(dec)

        return binv

    def op2bin(op):
        if op == "0":
            return "0000"
        elif op == ">":
            return "0000"
        elif op == ">=":
            return "0001"
        elif op == "<":
            return "0010"
        elif op == "<=":
            return "0011"
        elif op == "==":
            return "0100"
        elif op == "!=":
            return "0101"
        elif op == "+":
            return "1000"
        elif op == "-":
            return "1001"
        elif op == "*":
            return "1010"
        elif op == "&":
            return "0000"
        elif op == "|":
            return "0001"
        elif op == "^":
            return "0010"
        elif op == "~":
            return "0011"
        elif op == "<<":
            return "0100"
        elif op == ">>":
            return "0101"
        elif op == "upper":
            return "1010"
        elif op == "lower":
            return "0101"
        else:
            print("Error: operation \"%s\" not recognized" % op)
            return "1111"

    ######################################
    ### First pass: parse instructions ###
    ######################################
    for e in progList:
        inst = progList[e]['inst']
        args = progList[e]['args']

        # I-type: three registers and an immediate value.
        # I-type:<inst>:page:channel:oper:ra:rb:rc:imm

        # pushi p, $ra, $rb, imm
        if inst == 'pushi':
            comp_re = "\s*(\d+)\s*,\s*\$(\d+)\s*,\s*\$(\d+)\s*,\s*(\-?\d+)"
            m = re.search(comp_re, args)

            # If there is a match.
            if m:
                page = m.group(1)
                ra = m.group(2)
                rb = m.group(3)
                imm = m.group(4)

                # Add entry into structure.
                progList[e]['inst_parse'] = "I-type:pushi:" + \
                    page + ":0:0:" + rb + ":" + ra + ":0:" + imm

            # Error: bad instruction format.
            else:
                print("Error: bad format on instruction @%d: %s" % (e, inst))

        # popi p, $r
        elif inst == 'popi':
            comp_re = "\s*(\d+)\s*,\s*\$(\d+)"
            m = re.search(comp_re, args)

            # If there is a match.
            if m:
                page = m.group(1)
                r = m.group(2)

                # Add entry into structure.
                progList[e]['inst_parse'] = "I-type:popi:" + \
                    page + ":0:0:" + r + ":0:0:0"

            # Error: bad instruction format.
            else:
                print("Error: bad format on instruction @%d: %s" % (e, inst))

        # mathi p, $ra, $rb oper imm
        if inst == 'mathi':
            comp_re = "\s*(\d+)\s*,\s*\$(\d+)\s*,\s*\$(\d+)\s*([\+\-\*])\s*(0?x?\-?[0-9a-fA-F]+)"
            m = re.search(comp_re, args)

            # If there is a match.
            if m:
                page = m.group(1)
                ra = m.group(2)
                rb = m.group(3)
                oper = m.group(4)
                imm = m.group(5)

                # Add entry into structure.
                progList[e]['inst_parse'] = "I-type:mathi:" + page + \
                    ":0:" + oper + ":" + ra + ":" + rb + ":0:" + imm

            # Error: bad instruction format.
            else:
                print("Error: bad format on instruction @%d: %s" % (e, inst))

        # seti ch, p, $r, t
        if inst == 'seti':
            comp_re = "\s*(\d+)\s*,\s*(\d+)\s*,\s*\$(\d+)\s*,\s*(\-?\d+)"
            m = re.search(comp_re, args)

            # If there is a match.
            if m:
                ch = m.group(1)
                page = m.group(2)
                ra = m.group(3)
                t = m.group(4)

                # Add entry into structure.
                progList[e]['inst_parse'] = "I-type:seti:" + \
                    page + ":" + ch + ":0:0:" + ra + ":0:" + t

            # Error: bad instruction format.
            else:
                print("Error: bad format on instruction @%d: %s" % (e, inst))

        # synci t
        if inst == 'synci':
            comp_re = "\s*(\d+)"
            m = re.search(comp_re, args)

            # If there is a match.
            if m:
                t = m.group(1)

                # Add entry into structure.
                progList[e]['inst_parse'] = "I-type:synci:0:0:0:0:0:0:" + t

            # Error: bad instruction format.
            else:
                print("Error: bad format on instruction @%d: %s" % (e, inst))

        # waiti ch, t
        if inst == 'waiti':
            comp_re = "\s*(\d+)\s*,\s*(\d+)"
            m = re.search(comp_re, args)

            # If there is a match.
            if m:
                ch = m.group(1)
                t = m.group(2)

                # Add entry into structure.
                progList[e]['inst_parse'] = "I-type:waiti:0:" + \
                    ch + ":0:0:0:0:" + t

            # Error: bad instruction format.
            else:
                print("Error: bad format on instruction @%d: %s" % (e, inst))

        # bitwi p, $ra, $rb oper imm
        if inst == 'bitwi':
            comp_re = "\s*(\d+)\s*,\s*\$(\d+)\s*,\s*\$(\d+)\s*([&|<>^]+)\s*(0?x?\-?[0-9a-fA-F]+)"
            m = re.search(comp_re, args)

            # If there is a match.
            if m:
                page = m.group(1)
                ra = m.group(2)
                rb = m.group(3)
                oper = m.group(4)
                imm = m.group(5)

                # Add entry into structure.
                progList[e]['inst_parse'] = "I-type:bitwi:" + page + \
                    ":0:" + oper + ":" + ra + ":" + rb + ":0:" + imm

            # bitwi p, $ra, ~imm
            else:
                comp_re = "\s*(\d+)\s*,\s*\$(\d+)\s*,\s*~\s*(0?x?\-?[0-9a-fA-F]+)"
                m = re.search(comp_re, args)

                # If there is a match.
                if m:
                    page = m.group(1)
                    ra = m.group(2)
                    oper = "~"
                    imm = m.group(3)

                    # Add entry into structure.
                    progList[e]['inst_parse'] = "I-type:bitwi:" + \
                        page + ":0:" + oper + ":" + ra + ":0:0:" + imm

                # Error: bad instruction format.
                else:
                    print("Error: bad format on instruction @%d: %s" % (e, inst))

        # memri p, $r, imm
        if inst == 'memri':
            comp_re = "\s*(\d+)\s*,\s*\$(\d+)\s*,\s*(0?x?\-?[0-9a-fA-F]+)"
            m = re.search(comp_re, args)

            # If there is a match.
            if m:
                page = m.group(1)
                r = m.group(2)
                imm = m.group(3)

                # Add entry into structure.
                progList[e]['inst_parse'] = "I-type:memri:" + \
                    page + ":0:0:" + r + ":0:0:" + imm

            # Error: bad instruction format.
            else:
                print("Error: bad format on instruction @%d: %s" % (e, inst))

        # memwi p, $r, imm
        if inst == 'memwi':
            comp_re = "\s*(\d+)\s*,\s*\$(\d+)\s*,\s*(0?x?\-?[0-9a-fA-F]+)"
            m = re.search(comp_re, args)

            # If there is a match.
            if m:
                page = m.group(1)
                r = m.group(2)
                imm = m.group(3)

                # Add entry into structure.
                progList[e]['inst_parse'] = "I-type:memwi:" + \
                    page + ":0:0:0:0:" + r + ":" + imm

            # Error: bad instruction format.
            else:
                print("Error: bad format on instruction @%d: %s" % (e, inst))

        # regwi p, $r, imm
        if inst == 'regwi':
            comp_re = "\s*(\d+)\s*,\s*\$(\d+)\s*,\s*(0?x?\-?[0-9a-fA-F]+)"
            m = re.search(comp_re, args)

            # If there is a match.
            if m:
                page = m.group(1)
                r = m.group(2)
                imm = m.group(3)

                # Add entry into structure.
                progList[e]['inst_parse'] = "I-type:regwi:" + \
                    page + ":0:0:" + r + ":0:0:" + imm

            # Error: bad instruction format.
            else:
                print("Error: bad format on instruction @%d: %s" % (e, inst))

        # setbi ch, p, $r, t
        if inst == 'setbi':
            comp_re = "\s*(\d+)\s*,\s*(\d+)\s*,\s*\$(\d+)\s*,\s*(\-?\d+)"
            m = re.search(comp_re, args)

            # If there is a match.
            if m:
                ch = m.group(1)
                page = m.group(2)
                ra = m.group(3)
                t = m.group(4)

                # Add entry into structure.
                progList[e]['inst_parse'] = "I-type:setbi:" + \
                    page + ":" + ch + ":0:0:" + ra + ":0:" + t

            # Error: bad instruction format.
            else:
                print("Error: bad format on instruction @%d: %s" % (e, inst))

        # J-type: three registers and an address for jump.
        # J-type:<inst>:page:oper:ra:rb:rc:addr

        # loopnz p, $r, @label
        if inst == 'loopnz':
            comp_re = "\s*(\d+)\s*,\s*\$(\d+)\s*,\s*\@(.+)"
            m = re.search(comp_re, args)

            # If there is a match.
            if m:
                page = m.group(1)
                oper = "+"
                r = m.group(2)
                label = m.group(3)

                # Resolve symbol.
                if label in symbList:
                    label_addr = symbList[label]
                else:
                    print("Error: could not resolve symbol %s on instruction @%d: %s %s" % (
                        label, e, inst, args))

                # Add entry into structure.
                regs = r + ":" + r + ":0:" + str(label_addr)
                progList[e]['inst_parse'] = "J-type:loopnz:" + \
                    page + ":" + oper + ":" + regs

            # Error: bad instruction format.
            else:
                print("Error: bad format on instruction @%d: %s" % (e, inst))

        # condj p, $ra op $rb, @label
        if inst == 'condj':
            comp_re = "\s*(\d+)\s*,\s*\$(\d+)\s*([<>=!]+)\s*\$(\d+)\s*,\s*\@(.+)"
            m = re.search(comp_re, args)

            # If there is a match.
            if m:
                page = m.group(1)
                ra = m.group(2)
                oper = m.group(3)
                rb = m.group(4)
                label = m.group(5)

                # Resolve symbol.
                if label in symbList:
                    label_addr = symbList[label]
                else:
                    print("Error: could not resolve symbol %s on instruction @%d: %s %s" % (
                        label, e, inst, args))

                # Add entry into structure.
                regs = ra + ":" + rb + ":" + str(label_addr)
                progList[e]['inst_parse'] = "J-type:condj:" + \
                    page + ":" + oper + ":0:" + regs

            # Error: bad instruction format.
            else:
                print("Error: bad format on instruction @%d: %s" % (e, inst))

        # end
        if inst == 'end':
            # Add entry into structure.
            progList[e]['inst_parse'] = "J-type:end:0:0:0:0:0:0"

        # R-type: 8 registers, 7 for reading and 1 for writing.
        # R-type:<inst>:page:channel:oper:ra:rb:rc:rd:re:rf:rg:rh

        # math p, $ra, $rb oper $rc
        if inst == 'math':
            comp_re = "\s*(\d+)\s*,\s*\$(\d+)\s*,\s*\$(\d+)\s*([\+\-\*])\s*\$(\d+)"
            m = re.search(comp_re, args)

            # If there is a match.
            if m:
                page = m.group(1)
                ra = m.group(2)
                rb = m.group(3)
                oper = m.group(4)
                rc = m.group(5)

                # Add entry into structure.
                regs = ra + ":" + rb + ":" + rc + ":0:0:0:0:0"
                progList[e]['inst_parse'] = "R-type:math:" + \
                    page + ":0:" + oper + ":" + regs

            # Error: bad instruction format.
            else:
                print("Error: bad format on instruction @%d: %s" % (e, inst))

        # set ch, p, $ra, $rb, $rc, $rd, $re, $rt
        if inst == 'set':
            regs = "\s*\$(\d+)\s*,\s*\$(\d+)\s*,\s*\$(\d+)\s*,\s*\$(\d+)\s*,\s*\$(\d+)\s*,\s*\$(\d+)"
            comp_re = "\s*(\d+)\s*,\s*(\d+)\s*," + regs
            m = re.search(comp_re, args)

            # If there is a match.
            if m:
                ch = m.group(1)
                page = m.group(2)
                ra = m.group(3)
                rb = m.group(4)
                rc = m.group(5)
                rd = m.group(6)
                ree = m.group(7)
                rt = m.group(8)

                # Add entry into structure.
                regs = ra + ":" + rt + ":" + rb + ":" + rc + ":" + rd + ":" + ree + ":0"
                progList[e]['inst_parse'] = "R-type:set:" + \
                    page + ":" + ch + ":0:0:" + regs

            # Error: bad instruction format.
            else:
                print("Error: bad format on instruction @%d: %s" % (e, inst))

        # sync p, $r
        if inst == 'sync':
            comp_re = "\s*(\d+)\s*,\s*\$(\d+)"
            m = re.search(comp_re, args)

            # If there is a match.
            if m:
                page = m.group(1)
                r = m.group(2)

                # Add entry into structure.
                progList[e]['inst_parse'] = "R-type:sync:" + \
                    page + ":0:0:0:0:" + r + ":0:0:0:0:0"

            # Error: bad instruction format.
            else:
                print("Error: bad format on instruction @%d: %s" % (e, inst))

        # read ch, p, oper $r
        if inst == 'read':
            comp_re = "\s*(\d+)\s*,\s*(\d+)\s*,\s*(upper|lower)\s+\$(\d+)"
            m = re.search(comp_re, args)

            # If there is a match.
            if m:
                ch = m.group(1)
                page = m.group(2)
                oper = m.group(3)
                r = m.group(4)

                # Add entry into structure.
                progList[e]['inst_parse'] = "R-type:read:" + page + \
                    ":" + ch + ":" + oper + ":" + r + ":0:0:0:0:0:0:0"

            # Error: bad instruction format.
            else:
                print("Error: bad format on instruction @%d: %s" % (e, inst))

        # wait ch, p, $r
        if inst == 'wait':
            comp_re = "\s*(\d+)\s*,\s*(\d+)\s*,\s*\$(\d+)"
            m = re.search(comp_re, args)

            # If there is a match.
            if m:
                ch = m.group(1)
                page = m.group(2)
                r = m.group(3)

                # Add entry into structure.
                progList[e]['inst_parse'] = "R-type:wait:" + \
                    page + ":" + ch + ":0:0:0:" + r + ":0:0:0:0:0"

            # Error: bad instruction format.
            else:
                print("Error: bad format on instruction @%d: %s" % (e, inst))

        # bitw p, $ra, $rb oper $rc
        if inst == 'bitw':
            comp_re = "\s*(\d+)\s*,\s*\$(\d+)\s*,\s*\$(\d+)\s*([&|<>^]+)\s*\$(\d+)"
            m = re.search(comp_re, args)

            # If there is a match.
            if m:
                page = m.group(1)
                ra = m.group(2)
                rb = m.group(3)
                oper = m.group(4)
                rc = m.group(5)

                # Add entry into structure.
                regs = ra + ":" + rb + ":" + rc + ":0:0:0:0:0"
                progList[e]['inst_parse'] = "R-type:bitw:" + \
                    page + ":0:" + oper + ":" + regs

            # bitw p, $ra, ~$rb
            else:
                comp_re = "\s*(\d+)\s*,\s*\$(\d+)\s*,\s*~\s*\$(\d+)"
                m = re.search(comp_re, args)

                # If there is a match.
                if m:
                    page = m.group(1)
                    ra = m.group(2)
                    rb = m.group(3)
                    oper = "~"

                    # Add entry into structure.
                    regs = ra + ":0:" + ":" + rb + ":0:0:0:0:0"
                    progList[e]['inst_parse'] = "R-type:bitw:" + \
                        page + ":0:" + oper + ":" + regs

                # Error: bad instruction format.
                else:
                    print("Error: bad format on instruction @%d: %s" % (e, inst))

        # memr p, $ra, $rb
        if inst == 'memr':
            comp_re = "\s*(\d+)\s*,\s*\$(\d+)\s*,\s*\$(\d+)"
            m = re.search(comp_re, args)

            # If there is a match.
            if m:
                page = m.group(1)
                ra = m.group(2)
                rb = m.group(3)

                # Add entry into structure.
                regs = ra + ":" + rb + ":0:0:0:0:0:0"
                progList[e]['inst_parse'] = "R-type:memr:" + \
                    page + ":0:0:" + regs

            # Error: bad instruction format.
            else:
                print("Error: bad format on instruction @%d: %s" % (e, inst))

        # memw p, $ra, $rb
        if inst == 'memw':
            comp_re = "\s*(\d+)\s*,\s*\$(\d+)\s*,\s*\$(\d+)"
            m = re.search(comp_re, args)

            # If there is a match.
            if m:
                page = m.group(1)
                ra = m.group(2)
                rb = m.group(3)

                # Add entry into structure.
                regs = rb + ":" + ra + ":0:0:0:0:0"
                progList[e]['inst_parse'] = "R-type:memw:" + \
                    page + ":0:0:0:" + regs

            # Error: bad instruction format.
            else:
                print("Error: bad format on instruction @%d: %s" % (e, inst))

        # setb ch, p, $ra, $rb, $rc, $rd, $re, $rt
        if inst == 'setb':
            regs = "\s*\$(\d+)\s*,\s*\$(\d+)\s*,\s*\$(\d+)\s*,\s*\$(\d+)\s*,\s*\$(\d+)\s*,\s*\$(\d+)"
            comp_re = "\s*(\d+)\s*,\s*(\d+)\s*," + regs
            m = re.search(comp_re, args)

            # If there is a match.
            if m:
                ch = m.group(1)
                page = m.group(2)
                ra = m.group(3)
                rb = m.group(4)
                rc = m.group(5)
                rd = m.group(6)
                ree = m.group(7)
                rt = m.group(8)

                # Add entry into structure.
                regs = ra + ":" + rt + ":" + rb + ":" + rc + ":" + rd + ":" + ree + ":0"
                progList[e]['inst_parse'] = "R-type:setb:" + \
                    page + ":" + ch + ":0:0:" + regs

            # Error: bad instruction format.
            else:
                print("Error: bad format on instruction @%d: %s" % (e, inst))

    ######################################
    ### Second pass: convert to binary ###
    ######################################
    for e in progList:
        inst = progList[e]['inst_parse']
        spl = inst.split(":")

        # I-type
        if spl[0] == "I-type":
            # Instruction.
            if spl[1] in instList:
                inst_bin = instList[spl[1]]['bin']
            else:
                print(
                    "Error: instruction %s not found on instraction list" % spl[1])

            # page.
            page = unsigned2bin(spl[2], 3)

            # channel
            ch = unsigned2bin(spl[3], 3)

            # oper
            oper = op2bin(spl[4])

            # Registers.
            ra = unsigned2bin(spl[5], 5)
            rb = unsigned2bin(spl[6], 5)
            rc = unsigned2bin(spl[7], 5)

            # Immediate.
            imm = integer2bin(spl[8], 31)

            # Machine code (bin/hex).
            code = inst_bin + page + ch + oper + ra + rb + rc + imm
            code_h = "{:016x}".format(int(code, 2))

            # Write values back into hash.
            progList[e]['inst_bin'] = code
            progList[e]['inst_hex'] = code_h

        elif (spl[0] == "J-type"):
            # Instruction.
            if spl[1] in instList:
                inst_bin = instList[spl[1]]['bin']
            else:
                print(
                    "Error: instruction %s not found on instraction list" % spl[1])

            # Page.
            page = unsigned2bin(spl[2], 3)

            # Zeros.
            z3 = unsigned2bin("0", 3)

            # oper
            oper = op2bin(spl[3])

            # Registers.
            ra = unsigned2bin(spl[4], 5)
            rb = unsigned2bin(spl[5], 5)
            rc = unsigned2bin(spl[6], 5)

            # Zeros.
            z15 = unsigned2bin("0", 15)

            # Address.
            jmp_addr = unsigned2bin(spl[7], 16)

            # Machine code (bin/hex).
            code = inst_bin + page + z3 + oper + ra + rb + rc + z15 + jmp_addr
            code_h = "{:016x}".format(int(code, 2))

            # Write values back into hash.
            progList[e]['inst_bin'] = code
            progList[e]['inst_hex'] = code_h

        elif (spl[0] == "R-type"):
            # Instruction.
            if spl[1] in instList:
                inst_bin = instList[spl[1]]['bin']
            else:
                print(
                    "Error: instruction \"%s\" not found on instraction list" % spl[1])

            # Page.
            page = unsigned2bin(spl[2], 3)

            # Channel
            ch = unsigned2bin(spl[3], 3)

            # Oper
            oper = op2bin(spl[4])

            # Registers.
            ra = unsigned2bin(spl[5], 5)
            rb = unsigned2bin(spl[6], 5)
            rc = unsigned2bin(spl[7], 5)
            rd = unsigned2bin(spl[8], 5)
            ree = unsigned2bin(spl[9], 5)
            rf = unsigned2bin(spl[10], 5)
            rg = unsigned2bin(spl[11], 5)
            rh = unsigned2bin(spl[12], 5)

            # Zeros.
            z6 = unsigned2bin("0", 6)

            # Machine code (bin/hex).
            code = inst_bin + page + ch + oper + ra + \
                rb + rc + rd + ree + rf + rg + rh + z6
            code_h = "{:016x}".format(int(code, 2))

            # Write values back into hash.
            progList[e]['inst_bin'] = code
            progList[e]['inst_hex'] = code_h

        else:
            print("Error: bad type on instruction @%d: %s" % (e, inst))

    ####################
    ### Write output ###
    ####################
    # Binary format.
    if outfmt == "bin":
        for e in progList:
            outProg[e] = progList[e]['inst_bin']

    # Hex format.
    elif outfmt == "hex":
        for e in progList:
            out = progList[e]['inst_hex'] + " -> " + \
                progList[e]['inst'] + " " + progList[e]['args']
            outProg[e] = out

    else:
        print("Error: \"%s\" is not a recognized output format" % outfmt)

    # Return program list.
    return outProg


def parse_to_bin(path):
    """
    Parses the .asm assembly language tProc program into a form appropriate for QickSoc.load_bin_program().

    :param file: ASM program file name
    :type file: str
    :return: Program as a list of 64-bit ints
    :rtype: list
    """
    p = parse_prog(path)
    return [int(p[i], 2) for i in p]

def load_program(soc, prog="prog.asm", fmt="asm"):
    """
    Loads tProc program. If asm program, it compiles first

    :param soc: Qick to be programmed
    :type soc: QickSoc
    :param prog: program file name
    :type prog: string
    :param fmt: file format
    :type fmt: string
    """
    # Binary file format.
    if fmt == "bin":
        # Read binary file from disk.
        with open(prog, "r") as fd:
            progList = [int(line, 2) for line in fd]

    # Asm file.
    elif fmt == "asm":
        # Compile program.
        progList = parse_to_bin(prog)

    soc.load_bin_program(progList)



from threading import Thread, Event
from queue import Queue
import time
import numpy as np
import traceback

# This code originally used Process not Thread.
# Process is much slower to start (Process.start() is ~100 ms, Thread.start() is a few ms)
# The process-safe versions of Queue and Event are also significantly slower.
# On the other hand, CPU-bound Python threads can't run in parallel ("global interpreter lock").
# The overall problem is not CPU-bound - we should always be limited by tProc execution.
# In the worst case where the tProc is running fast, we should actually be waiting for IO a lot (due to the DMA).
# So we think it's safe to use threads.
# However, this is a complicated problem and we may ultimately need to mess around with sys.setswitchinterval() or go back to Process.
# To use Process instead of Thread, use the following import and change WORKERTYPE.
#from multiprocessing import Process, Queue, Event

class DataStreamer():
    """
    Uses a separate thread to read data from the average buffers.
    The class methods define the readout loop and initialization of the worker thread.
    The QickSoc methods start_readout() and poll_data() are the external interface to the streamer.

    We don't lock the QickSoc or the IPs. The user is responsible for not disrupting a readout in progress.

    :param soc: The QickSoc object.
    :type soc: QickSoc
    """

    #WORKERTYPE = Process
    WORKERTYPE = Thread

    def __init__(self, soc):
        self.soc = soc

        self.start_worker()

    def start_worker(self):
        # Initialize flags and queues.
        # Passes run commands from the main thread to the worker thread.
        self.job_queue = Queue()
        # Passes data from the worker thread to the main thread.
        self.data_queue = Queue()
        # Passes exceptions from the worker thread to the main thread.
        self.error_queue = Queue()
        # The main thread can use this flag to tell the worker thread to stop.
        # The main thread clears the flag when starting readout.
        self.stop_flag = Event()
        # The worker thread uses this to tell the main thread when it's done.
        # The main thread clears the flag when starting readout.
        self.done_flag = Event()
        self.done_flag.set()

        # Process object for the streaming readout.
        # daemon=True means the readout thread will be killed if the parent is killed
        self.readout_worker = self.WORKERTYPE(target=self._run_readout, daemon=True)
        self.readout_worker.start()

    def stop_readout(self):
        """
        Signal the readout loop to break.
        """
        self.stop_flag.set()

    def readout_running(self):
        """
        Test if the readout loop is running.

        :return: readout thread status
        :rtype: bool
        """
        return not self.done_flag.is_set()

    def data_available(self):
        """
        Test if data is available in the queue.

        :return: data queue status
        :rtype: bool
        """
        return not self.data_queue.empty()

    def _run_readout(self):
        """
        Worker thread for the streaming readout

        :param total_count: Number of data points expected
        :type addr: int
        :param counter_addr: Data memory address for the loop counter
        :type counter_addr: int
        :param ch_list: List of readout channels
        :type addr: list of int
        :param reads_per_count: Number of data points to expect per counter increment
        :type reads_per_count: list of int
        """
        while True:
            try:
                # wait for a job
                total_shots, counter_addr, ch_list, reads_per_count, stride = self.job_queue.get(block=True)
                #print("streamer loop: start", total_count)

                shots = 0
                last_shots = 0

                # how many shots worth of data to transfer at a time
                if stride is None:
                    stride = int(0.1 * self.soc.get_avg_max_length(0)/max(reads_per_count))
                # bigger stride is more efficient, but the transfer size must never exceed AVG_MAX_LENGTH, so the stride should be set with some safety margin

                # make sure count variable is reset to 0 before starting processor
                self.soc.set_tproc_counter(addr=counter_addr, val=0)
                stats = []

                t_start = time.time()

                # if the tproc is configured for internal start, this will start the program
                # for external start, the program will not start until a start pulse is received
                self.soc.start_tproc()

                # Keep streaming data until you get all of it
                while last_shots < total_shots:
                    if self.stop_flag.is_set():
                        print("streamer loop: got stop flag")
                        break
                    shots = self.soc.get_tproc_counter(addr=counter_addr)
                    # wait until either you've gotten a full stride of measurements or you've finished (so you don't go crazy trying to download every measurement)
                    if shots >= min(last_shots+stride, total_shots):
                        newshots = shots-last_shots
                        # buffer for each channel
                        d_buf = [None for nreads in reads_per_count]

                        # for each adc channel get the single shot data and add it to the buffer
                        for iCh, ch in enumerate(ch_list):
                            newpoints = newshots*reads_per_count[iCh]
                            if newpoints >= self.soc.get_avg_max_length(ch):
                                raise RuntimeError("Overflowed the averages buffer (%d unread samples >= buffer size %d)."
                                                   % (newpoints, self.soc.get_avg_max_length(ch)) +
                                                   "\nYou need to slow down the tProc by increasing relax_delay." +
                                                   "\nIf the TQDM progress bar is enabled, disabling it may help.")

                            addr = last_shots * reads_per_count[iCh] % self.soc.get_avg_max_length(ch)
                            data = self.soc.get_accumulated(ch=ch, address=addr, length=newpoints)
                            d_buf[iCh] = data

                        last_shots += newshots

                        stats = (time.time()-t_start, shots, addr, newshots)
                        self.data_queue.put((newshots, (d_buf, stats)))
                #if last_count==total_count: print("streamer loop: normal completion")

            except Exception as e:
                print("streamer loop: got exception")
                # traceback.print_exc()
                # pass the exception to the main thread
                self.error_queue.put(e)
                # put dummy data in the data queue, to trigger a poll_data read
                self.data_queue.put((0, (None, None)))
            finally:
                # we should set the done flag regardless of whether we completed readout, used the stop flag, or errored out
                self.done_flag.set()
                # set tproc for internal start so we don't run the program repeatedly (this also clears the internal-start register)
                self.soc.start_src("internal")
